#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['pypants', 'pypants.build_targets', 'pypants.generators']

package_data = \
{'': ['*']}

install_requires = \
['astor<1,>=0.7.1',
 'black==19.3b0',
 'Click>=7.0',
 'cookiecutter<2,>=1.6.0',
 'networkx<3,>=2.2',
 'python-slugify<2,>=1.2.0']

entry_points = \
{'console_scripts': ['pypants = pypants.cli:cli']}

setup(name='pypants',
      version='0.1.4',
      description='CLI for working with Python packages and BUILD files in a Pants monorepo',
      author='Jonathan Drake',
      author_email='jdrake@narrativescience.com',
      url='https://github.com/NarrativeScience/pypants',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      entry_points=entry_points,
      python_requires='>=3.6,<4',
     )
